from itertools import chain
from operator import attrgetter


__all__ = ("Contents", "contents_for_items", "contents_for_item")


class Contents:
    def __init__(self, regions):
        self.regions = regions
        self._sorted = False
        self._contents = {region.key: [] for region in self.regions}
        self._unknown_region_contents = []

    def add(self, content):
        self._sorted = False
        try:
            self._contents[content.region].append(content)
        except KeyError:
            self._unknown_region_contents.append(content)

    def _sort(self):
        for region_key in list(self._contents):
            self._contents[region_key] = sorted(
                self._contents[region_key], key=attrgetter("ordering")
            )
        self._sorted = True

    def __getattr__(self, key):
        if key.startswith("_"):
            raise AttributeError(f"Invalid region key {key!r} on {self!r}")
        if not self._sorted:
            self._sort()
        return self._contents.get(key, [])

    def __getitem__(self, key):
        if key.startswith("_"):
            raise KeyError(f"Invalid region key {key!r} on {self!r}")
        if not self._sorted:
            self._sort()
        return self._contents.get(key, [])

    def __iter__(self):
        if not self._sorted:
            self._sort()
        return chain.from_iterable(
            self._contents[region.key] for region in self.regions
        )

    def __len__(self):
        return sum((len(contents) for contents in self._contents.values()), 0)

    def inherit_regions(self, contents):
        for region in self.regions:
            if not region.inherited or self[region.key]:
                continue
            self._contents[region.key] = contents[region.key]  # Still sorted


def contents_for_items(items, plugins, *, regions=None):
    contents = {item: Contents(regions or item.regions) for item in items}
    items_dict = {item.pk: item for item in contents}
    for plugin in plugins:
        queryset = plugin.get_queryset().filter(parent__in=contents.keys())
        if regions is not None:
            queryset = queryset.filter(region__in=[region.key for region in regions])
        queryset._known_related_objects.setdefault(
            plugin._meta.get_field("parent"), {}
        ).update(items_dict)
        for obj in queryset:
            contents[obj.parent].add(obj)
    return contents


def contents_for_item(item, plugins, *, inherit_from=None, regions=None):
    inherit_from = list(inherit_from) if inherit_from else []
    all_contents = contents_for_items(
        [item] + inherit_from, plugins=plugins, regions=regions
    )
    contents = all_contents[item]
    for other in inherit_from:
        contents.inherit_regions(all_contents[other])
    return contents



from django.core.exceptions import ImproperlyConfigured
from django.db import models


__all__ = ("Type", "Region", "Template", "create_plugin_base")


class Type(dict):
    _REQUIRED = set()

    def __init__(self, **kwargs):
        missing = self._REQUIRED - set(kwargs)
        if missing:
            raise TypeError(
                f"Missing arguments to {self.__class__.__name__}: {missing}"
            )
        super().__init__(**kwargs)

    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError as exc:
            raise AttributeError(f"Unknown attribute {attr!r}") from exc


class Region(Type):
    _REQUIRED = {"key", "title", "inherited"}

    def __init__(self, *, key, **kwargs):
        kwargs.setdefault("inherited", False)
        if key == "regions":
            raise ImproperlyConfigured("'regions' cannot be used as a Region key.")
        elif key.startswith("_"):
            raise ImproperlyConfigured(f"Region key {key!r} cannot start with '_'.")
        elif not key.isidentifier():
            raise ImproperlyConfigured(f"Region key {key!r} is no identifier.")
        super().__init__(key=key, **kwargs)


class Template(Type):
    _REQUIRED = {"key", "template_name", "title", "regions"}


def create_plugin_base(content_base):
    """
    Create and return a base class for plugins

    The base class contains a ``parent`` foreign key and the required
    ``region`` and ``ordering`` fields.
    """

    class PluginBase(models.Model):
        parent = models.ForeignKey(
            content_base,
            related_name="%(app_label)s_%(class)s_set",
            on_delete=models.CASCADE,
        )
        region = models.CharField(max_length=255)
        ordering = models.IntegerField(default=0)

        class Meta:
            abstract = True
            app_label = content_base._meta.app_label
            ordering = ["ordering"]

        def __str__(self):
            return f"{self._meta.label}<region={self.region} ordering={self.ordering} pk={self.pk}>"  # pragma: no cover

        @classmethod
        def get_queryset(cls):
            return cls.objects.all()

    return PluginBase
